package FDN::Sql::Structure;

use strict;

use Encode;
use Data::Dumper;
use FDN::Sql::DB_Access;
use FDN::Sql::Basic;
use FDN::Sql::DBC;
use FDN::Common::Config;
use FDN::Common::Modules;
use FDN::Common::Logger;

use base 'Exporter';

our @EXPORT = qw (
	switch_option
	add_sql_file
	set_db_name
	set_db_force
	set_login
	set_passwd
	set_utf8
	set_type
	set_host
	set_source
	set_config
	set_config_file
	control_options
	check_db
	$subst
);

my $db_name;
my ( $login, $passwd, $utf8, $host, $source ) = ( undef, 'toto', undef, 'localhost', '%' );

my @files;
my @actions;
my $dbh;
my $create;
my $subst = {};

my $config_file = 'check-db.cfg';
my $debug = 0;
my $force = 0;
my $dry_alter = 1;
my $dry_db    = 1;
my $dry_table = 1;
my $dry_data  = 1;
my $dry_cfg   = 1;


# The new definition (target), will not change during process.
my @queries = ();
my $tables_new = {};
my $limits = {
	'pre' => [],
	'pre-del' => [],
	'pre-create' => [],
	'post-del' => [],
	'post' => []
};






my $appli_id_par_name;
my $text_par_xid_appli_id;
my $create_text;
my $set_text_simple;
my $set_text_val;
my $new_appli;
my $main_domain;

# HERE

my $dir; # Should become deprecated
my $source_cr;
my $source_db;
my $loc_cr;
my $loc_db;
my $applis;
my $connect;
my $data;
my $texts;



sub __debug($) {
	my ( $msg ) = @_;
	debug $msg if $debug;
}


#############################################################################################
# To build the scripts
#############################################################################################


sub get_base_dir() {
	my $curdir = `/bin/pwd`;
	chomp $curdir;
	$curdir =~ s/([^\/])$/$1\//;
	$curdir =~ s"^.*/GCMP/"./";
	$curdir =~ s"/[^/]*/"/../"g;
	return $curdir;
}

sub get_main_domain() {
	$main_domain = Config_Key('/etc/gcmp/common.cfg','domains','main');
	die "I need to find main domain in /etc/gcmp/common.cfg" if ! defined $main_domain;
}

sub slurp_file($) {
	my ( $file ) = @_;
	local $/;
	open IN, "<$file" || die "Cannot read $file: $!";
	my $text = <IN>;
	close IN;
	return $text;
}

sub read_files() {
	for my $src ( @files ) {
		print "Reading from file $src\n";
		$create .= slurp_file($src);
	}
	my ( $k, $v );
	while ( ($k,$v) = each %$subst ) {
		$create =~ s/$k/$v/sg;
	}
	@queries = split /;\s*$/m, $create;
}

#############################################################################################
# Some utilities...
#############################################################################################

sub run_or_die($$$) {
	my ( $dbh, $type, $req ) = @_;
	my $tmp = sql_transaction($dbh, $type, $req);
	if ( ! $tmp->{'ok'} ) {
		die "Failed with query $req\n".Dumper($tmp);
	}
	return $tmp;
}

sub dry_or_run($$$$$) {
	my ( $dry, $text, $dbh, $type, $req ) = @_;
	if ( $dry ) {
		print "-- ".$text if defined $text;
		print $req.";\n";
		return;
	} else {
		print $text if defined $text;
	}
	run_or_die($dbh, $type, $req);
}

sub print_err($$) {
	my ( $err, $msg ) = @_;
	$$err = 1;
	print 'ERR '.$msg;
}

##############################################################################################
# On the 'config' action
##############################################################################################


sub config() {
	my $cfg;
	my $add;
	my $err;
	if ( ! -f '/etc/fdn/dbc.cfg' ) {
		$cfg = {};
		$add = "";
	} else {
		$cfg = Load_Config('dbc.cfg');
	}

	if ( exists $cfg->{$db_name} ) {
		print "Configuration for $db_name already exists.\n";
		$err .= "Wrong type for $db_name, should be 'mysql'\n" if ( $cfg->{$db_name}->{'type'} ne 'mysql' );
		$err .= "Wrong host for $db_name, is '$cfg->{$db_name}->{'host'}' should be '$host'\n" if ( $cfg->{$db_name}->{'host'} ne $host );
		$err .= "Wrong name for $db_name\n" if ( $cfg->{$db_name}->{'name'} ne $db_name );
#		$connect->{$db}->[0] = $cfg->{$db}->{'user'};
#		$connect->{$db}->[1] = $cfg->{$db}->{'passwd'};
#		$connect->{$db}->[2] = $cfg->{$db}->{'utf8'};
	} else {
		print "Need to add $db_name in config file.\n";
		$add .= "[$db_name]\n".
			"\ttype\t= mysql\n".
			"\thost\t= $host\n".
			"\tname\t= $db_name\n".
			"\tuser\t= ".$login."\n".
			"\tpasswd\t= ".$passwd."\n";
		$add .= "\tutf8\t= ".$utf8."\n" if defined $utf8;
		$add .= "\n";
	}
	if ( $err ) {
		die "Database $db_name is wrongly configured in /etc/fdn/dbc.cfg config file:\n$err\n";
	}
	if ( $add ) {
		if ( $dry_cfg && ! $force ) {
			print "Need to update config file /etc/fdn/dbc.cfg by adding:\n".$add;
		} else {
			print "Updating config file\n";
			Forget_Config('dbc.cfg');
			open OUT, ">>/etc/fdn/dbc.cfg" or die "Cannot update /etc/fdn/dbc.cfg: $!\n";
			print OUT $add;
			close OUT;
		}
	}
}

##############################################################################################
# On the 'create' action
##############################################################################################

sub create() {
	my $dbh = do_connect('GLOBAL');
	if ( ! $dbh ) {
		die "Failed to connect to mysql as root\n";
	}
	my $tmp = run_or_die($dbh,'show','show databases');
	for my $r ( @{$tmp->{'rows'}} ) {
		my $db = $r->{'Database'};
		if ( $db eq $db_name ) {
			print "Database already exists\n";
			return;
		}
	}

	print "-- Need to create database $db_name\n";
	my ( $rq1, $rq2, $rq3, $rq4 );
	$rq1 = "create database $db_name";
	#$rq2 = 'grant all privileges on * to \''.$login.'\'@\''.$source.'\' identified by \''.$passwd.'\'';
	$rq3 = 'grant all privileges on '.$db_name.'.* to \''.$login.'\'@\''.$source.'\' identified by \''.$passwd.'\'';
	$rq4 = 'flush privileges';
	if ( $dry_db && ! $force ) {
		print $rq1.";\n";
		#print $rq2.";\n";
		print $rq3.";\n";
		print $rq4.";\n";
	} else {
		run_or_die($dbh,'create',$rq1);
		#run_or_die($dbh,'grant',$rq2);
		run_or_die($dbh,'grant',$rq3);
		run_or_die($dbh,'flush',$rq4);
	}
	disconnect_db($dbh);
}

##############################################################################################
# About the structure of the database
##############################################################################################


sub read_table_structure($) {
	my ( $sql ) = @_;
	my $res = {};
	$sql =~ s/\s*--.*$//gm;
	die "Cannot understand <$sql>" if ( $sql !~ /create\s+table\s+`?([a-z0-9_]+)`?\s*\((.*)\)\s*(\S.*\S)?\s*$/is );
	my ( $table, $columns, $options )  = ( $1, $2, $3 );
	$res->{'sql'} = $sql;
	$res->{'cols'} = { list => []};
	my $err = 0;
	while ( $columns =~ s/^\s*
				`?(\w+)`?
			       \s+
				(  (?: bigint | int | integer | tinyint )(?:\s*\(\d+\))?  (?:\s+unsigned)?
				 | (?: char | varchar         )(?:\s*\(\d+\))   (?:\s+character\s+set\s+\S+)?
				 | text                                         (?:\s+character\s+set\s+\S+)?
				 | enum\s*\([^)]+\)                             (?:\s+character\s+set\s+\S+)?
				 | varbinary\(\d+\)
				 | boolean | bool
				 | datetime | date 
				 | timestamp(?:\s*\(\d+\))?
				)
			       \s*
				(  (?:unique\s+)?
				   (?: null | (?:not\s+null)?  (?:\s*default\s+(?:\S+|'[^']*')(?:\s+on\s+update\s+\S+)?)? | default\s+\S+\s+not\s+null )
				)
			       \s*
				(,|$)//isx ) {
		my ($col, $type, $dflt) = ($1, $2, $3);
		push @{$res->{'cols'}->{'list'}}, $col;
		$res->{'cols'}->{$col} = { type => $type, dflt => $dflt };
		print_err(\$err,"not null and default in wrong order, in $table.$col\n") if ( $dflt =~ /default\s+\S+\s+not\s+null/i );
	}
	$res->{'idxs'} = {};
	while ( $columns =~ s/^\s*
				( primary\s+key | index | unique(?:\s+index)? | key | unique\skey )
			       \s*
				(`\w+`)?
			       \s*
				(\( \s*`?\w+`?\s*(?:\(\s*\d+\s*\)\s*)? (?:, \s*`?\w+`?(?:\(\s*\d+\s*\)\s*)?)* \))
			       \s*
				(,|$)//isx ) {
		my ( $type, $name, $idx ) = ( $1, $2, $3 );
		$idx =~ s/`//g;
		$idx =~ s/\s//g;
		if ( defined $res->{'idxs'}->{$idx} ) {
			print_err(\$err,"There are several indexes about $idx in table $table\n");
		} else {
			$res->{'idxs'}->{$idx} = []
		}
		push @{$res->{'idxs'}->{$idx}}, { type => $type, name => $name };
	}
	if ( $columns !~ /^\s*$/is) {
		die "J'ai essay d'analyser a:\n$sql\n\nIl me reste a que je ne sais pas lire:\n<<<$columns>>>\n\n";
	}

	$res->{'opts'} = {};
	while ( $options =~ s/\s*(\S[^=]*\S)\s*=\s*(\S*)\s*// ) {
		my ( $opt, $val ) = ( $1, $2 );
		$opt = 'type' if ( $opt eq 'ENGINE' );
		$opt = 'default charset' if ( $opt eq 'DEFAULT CHARSET' );
		$res->{'opts'}->{$opt} = $val;
	}
	if ( $err ) {
		print "ERR J'ai vu des problmes:\n".$sql."\n".Dumper($res);
	}
	return $res;
}

sub get_base_type($) {
	my ( $type ) = @_;
	if ( $type =~ /tinyint|int|integer|bigint|bool|^enum/i ) {
		return 'int';
	} elsif ( $type =~ /timestamp/i ) {
		return 'ts';
	} elsif ( $type =~ /date/i ) {
		return 'date';
	} else {
		return 'text';
	}
}

sub drop_column($$$) {
	my ( $dbh, $table, $col ) = @_;
	my $req = "alter table $table drop column $col";
	my $redo = 0;
	if ( $dry_alter ) {
		print "-- Column $col should be dropped\n";
		print $req.";\n";
	} else {
		print "Dropping column $col from table $table\n";
		run_or_die($dbh, 'alter', $req);
		$redo = 1;
	}
	return $redo;
}

sub add_column($$$$$$) {
	my ( $dbh, $table, $col, $type, $default, $position ) = @_;
	my $req = "alter table $table add column $col $type $default $position";
	my $redo = 0;
	if ( $dry_alter ) {
		print "-- Column $col should be created $position\n";
		print $req.";\n";
		$redo = 1;
	} else {
		print "Adding column $col to table $table\n";
		run_or_die($dbh,'alter',$req);
		$redo = 1;
	}
	return $redo;
}

sub modify_column($$$$$) {
	my ( $dbh, $table, $col, $type, $default ) = @_;
	my $req = "alter table $table modify $col $type $default";
	my $redo = 0;
	if ( $dry_alter ) {
		print "-- Column type change for $table.$col, to $type\n";
		print $req.";\n";
	} else {
		print "Changing type of column $table.$col\n";
		run_or_die($dbh,'alter',$req);
		$redo = 1;
	}
	return $redo;
}

sub set_table_option($$$$) {
	my ( $dbh, $table, $name, $value ) = @_;
	my $req = "alter table $table ";
	$req .= 'ENGINE' if ( $name eq 'type' );
	$req .= 'default charset' if ( $name eq 'default charset' );
	$req .= " = ".$value;
	my $redo = 0;
	if ( $dry_alter ) {
		print "-- Option '$name' change on table $table, to $value\n";
		print $req.";\n";
	} else {
		print "Changing option $name on table $table\n";
		run_or_die($dbh, 'alter', $req);
		$redo = 1;
	}
	return $redo;
}

sub drop_index($$$$) {
	my ( $dbh, $table, $type, $name ) = @_;
	my $req;
	if ( $type =~ /^(unique )?(index|key)$/i ) {
		$req = "alter table $table drop index $name";
	} elsif ( $type =~ /^primary key$/i ) {
		$req = "alter table $table drop primary key";
	} else {
		die "Unknown column type $type";
	}
	if ( $dry_alter ) {
		print "-- Index $type/$name on $table should be dropped\n";
		print $req.";\n";
	} else {
		print "Dropping index $type/$name on table $table\n";
		run_or_die($dbh,'alter',$req);
	}
}

sub create_index($$$$) {
	my ( $dbh, $table, $type, $keys ) = @_;
	my $req;
	$req = "alter table $table add $type$keys";
	if ( $dry_alter ) {
		print "-- Index $type$keys should be created on table $table\n";
		print $req.";\n";
	} else {
		print "Creation index $type$keys on table $table\n";
		run_or_die($dbh,'alter',$req);
	}
}

sub compare_columns($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	my $redo = 0;
	for my $c ( @{$old->{'cols'}->{'list'}} ) {
		if ( ! exists $new->{'cols'}->{$c} ) {
			$redo += drop_column($dbh, $table, $c);
			delete $old->{'cols'}->{$c};
		}
	}
	my $prev;
	for my $c ( @{$new->{'cols'}->{'list'}} ) {
		if ( ! exists $old->{'cols'}->{$c} ) {
			my $type = $new->{'cols'}->{$c}->{'type'};
			my $dflt = $new->{'cols'}->{$c}->{'dflt'};
			my $pos;
			$pos = 'first' if ( ! $prev );
			$pos = 'after '.$prev if ( $prev );
			$redo += add_column($dbh, $table, $c, $type, $dflt, $pos);
			$old->{'cols'}->{$c} = $new->{'cols'}->{$c};
		}
		$prev = $c;
	}
	$old->{'cols'}->{'list'} = [ @{$new->{'cols'}->{'list'}} ];
	return $redo;
}

sub compare_column_types($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	my $old_charset = 'INVALID';
	if ( defined $old->{'opts'}->{'default charset'} ) {
		$old_charset = $old->{'opts'}->{'default charset'};
	}
	# Maintenant, on a les mme colonnes, vrifier les types.
	my $first_date = 1;
	my $redo = 0;
	for my $c ( @{$old->{'cols'}->{'list'}} ) {
		my $new_type = $new->{'cols'}->{$c}->{'type'};
		my $new_dflt = $new->{'cols'}->{$c}->{'dflt'};
		my $old_type = $old->{'cols'}->{$c}->{'type'};
		my $old_dflt = $old->{'cols'}->{$c}->{'dflt'};

		my $old_base = get_base_type($old_type);
		my $new_base = get_base_type($new_type);

		if ( $old_base ne $new_base ) {
			print "-- Need to change column type from $old_base to $new_base\n";
			$redo += modify_column($dbh, $table, $c, $new_type, $new_dflt);
			next;
		}

		my $re = $new_type;
		$re =~ s/\s*character\s+set\s+$old_charset\s*//;
		$re =~ s/\(/\\\(/g;
		$re =~ s/\)/\\\)/g;
		$re =~ s/',\s+'/','/g if ( $re =~ /enum/i );
		$re =~ s/\s+\\\(/\\s*\\\(/g;
		$re =~ s/\s+/\\s+/g;
		$re =~ s/integer\\s\+unsigned/int\\\(10\\\)\\s+unsigned/;
		$re =~ s/integer/int\\\(11\\\)/;
		$re =~ s/timestamp\\\(14\\\)/timestamp/;
		$re = 'tinyint\(1\)' if ( $re =~ /^\s*(bool|boolean)\s*$/i );
		if ( $old_type !~ /$re/i ) {
			print "-- Need to change column type from $old_type to $new_type\n";
			$redo += modify_column($dbh, $table, $c, $new_type, $new_dflt);
			next;
		}
		if ( $old_base eq $new_base ) {
			$re = $new_dflt;
			if ( $old_base eq 'ts' ) {
				if ( $first_date ) {
					$first_date = 0;
					if ( $re eq '' ) { $re = 'not\\s+null\\s+default\\s+CURRENT_TIMESTAMP\\s+on\\s+update\\s+CURRENT_TIMESTAMP'; } 
				} else {
					if ( $re eq '' ) { $re = 'not\\s+null\\s+default\\s+\'0000-00-00 00:00:00\''; }
				}
			}

			if ( $re eq '' ) { $re = '((default\s*)?null)?'; }
			if ( $re eq 'null') { $re = '(default\s|)null'; }
			if ( $re =~ /default/i ) {
				if ( $re !~ /not null/i ) {
					if ( $re !~ /null/i ) {
						$re = '(not\s+null\s+)?'.$re;
					}
				}
			}
			if ( $re !~ /default null/i ) {
				$re =~ s/default ('|)([^']*)\1(\s|$)/default '$2'/;
			}
			$re =~ s/'true'/'1'/;
			$re =~ s/'false'/'0'/;
			if ( $old_dflt !~ /^$re$/i ) {
				print "-- Need to change column type from $old_dflt to $new_dflt\n";
				$redo += modify_column($dbh, $table, $c, $new_type, $new_dflt );
			}
		}
	}
	return $redo;
}

sub compare_table_options($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	my $redo = 0;
	for my $o ( keys %{$old->{'opts'}} ) {
		if ( exists $new->{'opts'}->{$o} ) {
			$old->{'opts'}->{$o} =~ s/^('|)(.*)\1$/$2/;
			$new->{'opts'}->{$o} =~ s/^('|)(.*)\1$/$2/;
			if ( $old->{'opts'}->{$o} ne $new->{'opts'}->{$o} ) {
				$redo += set_table_option($dbh,$table,$o,$new->{'opts'}->{$o});
			}
		} else {
			if ( $dry_alter ) {
				print "-- Option '$o' deleted from table $table (was ".$old->{'opts'}->{$o}.")\n";
				print "-- Cannot remove an option from a table. CHECK.\n";
			} else {
				die "Cannot remove an option from a table (option $o from table $table)\n";
			}
		}
	}
	for my $o ( keys %{$new->{'opts'}} ) {
		if ( ! exists $old->{'opts'}->{$o} ) {
			$redo += set_table_option($dbh,$table,$o,$new->{'opts'}->{$o});
		}
	}
	return $redo;
}

sub compare_indices($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;
	for my $i ( keys %{$old->{'idxs'}} ) {
		my $old_type = $old->{'idxs'}->{$i}->[0]->{'type'};
		my $old_name = $old->{'idxs'}->{$i}->[0]->{'name'};
		my $re;
		if ( $old_type eq 'KEY' ) {
			$re = '(?:index|key)';
		} elsif ( $old_type eq 'PRIMARY KEY' ) {
			$re = 'primary key';
		} elsif ( $old_type eq 'UNIQUE KEY' ) {
			$re = 'unique(?:\s+index|\s+key|)';
		}
		if ( exists $new->{'idxs'}->{$i} ) {
			my $new_type = $new->{'idxs'}->{$i}->[0]->{'type'};
			if ( $new_type !~ /$re/i ) {
				drop_index($dbh, $table, $old_type, $old_name);
				create_index($dbh, $table, $new_type, $i);
			}
		} else {
			drop_index($dbh, $table, $old_type, $old_name);
		}
	}
	for my $i ( keys %{$new->{'idxs'}} ) {
		my $new_type = $new->{'idxs'}->{$i}->[0]->{'type'};
		if ( ! exists $old->{'idxs'}->{$i} ) {
			create_index($dbh, $table, $new_type, $i);
		}
	}
	return 0;
}

sub compare_tables($$$$) {
	my ( $dbh, $table, $old, $new ) = @_;

	my $redo = 0;
	$redo += compare_columns($dbh, $table, $old, $new);
	$redo += compare_column_types($dbh, $table, $old, $new);
	$redo += compare_table_options($dbh, $table, $old, $new);
	return 1 if ( $redo );
	compare_indices($dbh, $table, $old, $new);
	return 0;
}

sub get_old_tables_description($$) {
	my ( $dbh, $db ) = @_;
	my $tmp = run_or_die($dbh,'show','show tables');
	my $tables = {};
	for my $r ( @{$tmp->{'rows'}} ) {
		my $tbl = $r->{'Tables_in_'.$db};
		$tables->{$tbl} = 1;
	}
	for my $t ( keys %$tables ) {
		$tmp = get_table_create($dbh,$t);
		$tables->{$t} = read_table_structure('create table '.$t.' '.$tmp);
	}
	return $tables;
}

sub create_table($$$) {
	my ( $dbh, $table, $req ) = @_;
	if ( $dry_table ) {
		print "-- Table $table is new, need to create it.\n";
		print $req.";\n";
	} else {
		print "Creating new table $table\n";
		run_or_die($dbh,'create',$req);
	}
}

sub drop_table($$) {
	my ( $dbh, $table ) = @_;
	my $req = "drop table $table\n";
	if ( $dry_table ) {
		print "-- Table $table is not anymore used, need to drop it\n";
		print $req.";\n";
	} else {
		print "Dropping deprecated table $table\n";
		run_or_die($dbh,'drop',$req);
	}
}

sub read_limit_info($) {
	my ( $req ) = @_;
	my ( $type, $limit, $value, $script );
	my $res = {};
	if ( $req =~ s/-- limit (\S+) script (\S+) (\S+) (\S+)\s*$//m ) {
		( $type, $limit, $value, $script ) = ( $1, $2, $3, $4 );
	} else {
		die "There is a limit on the create sequence I cannot handle";
	}
	
	if ( $limit !~ /^(VERSION_NAME)$/ ) {
		die "There is a limit, but not about VERSION_NAME, I cannot handle";
	}
	if ( $value !~ /^\d+$/ ) {
		die "There is a limit on VERSION_NAME with non numerical version number";
	}
	if ( $script !~ /^\// ) {
		$script = '/usr/bin/'.$script;
	}
	if ( ! -x $script ) {
		die "There is a script I am supposed to run, $script, but I cannot find it";
	}
	$res->{'type'} = $type;
	$res->{'version'} = $value;
	$res->{'script'} = $script;
	return $res;
}

sub get_current_version($) {
	my ( $dbh ) = @_;
	my $tmp = sql_transaction($dbh,'select','select VERSION_NAME from VERSION');
	if ( ! $tmp->{'ok'} ) {
		return -1;
	}
	return $tmp->{'rows'}->[0]->{'VERSION_NAME'};
}

sub check_limits($$$) {	
	my ( $dbh, $limits, $curver ) = @_;
	for my $l ( @$limits ) {
		if ( $curver < $l->{'version'} ) {
			my $script = $l->{'script'};
			if ( $dry_table ) {
				print "-- I have to run $script\n";
			} else {
				print "Running ".$l->{'script'}."\n";
				system($l->{'script'}) == 0 or die "Failed to run script ".$l->{'script'};
			}
		}
	}
}

sub parse_db_structure() {
	for my $q ( @queries ) {
		while ( $q =~ /-- limit (pre|pre-del|pre-create|post-del|post) / ) {
			push @{$limits->{$1}}, read_limit_info($q);
			if ( $q !~ s/(-- limit .*)$//m ) {
				die "Pas pu retirer la limite dans $q\n";
			}
		}
		$q =~ s/^-- .*$//m;
		next if $q =~ /^\s*$/s;
		if ( $q =~ /create\s+table\s+([a-z0-9_]+)\W/is ) {
			$tables_new->{$1} = read_table_structure($q);
		}
	}
}

sub check_db_structure() {

	my $curver = get_current_version($dbh);

	my $redo = 1;
	my $nb_run = 0;
	my $tables_old;
	check_limits($dbh,$limits->{'pre'},$curver);
	while ( $redo && $nb_run < 2 ) {
		$nb_run++;
		$redo = 0;
		$tables_old = get_old_tables_description($dbh, $db_name);

		for my $t ( keys %$tables_new ) {
			if ( $tables_old->{$t} ) {
				$redo += compare_tables($dbh, $t, $tables_old->{$t}, $tables_new->{$t});
			}
		}
	}

	check_limits($dbh,$limits->{'pre-create'},$curver);
	for my $t ( keys %$tables_new ) {
		if ( ! $tables_old->{$t} ) {
			create_table($dbh, $t, $tables_new->{$t}->{'sql'});
		}
	}

	check_limits($dbh,$limits->{'pre-del'},$curver);
	for my $t ( keys %$tables_old ) {
		if ( ! $tables_new->{$t} ) {
			drop_table($dbh, $t);
		}
	}
	check_limits($dbh,$limits->{'post-del'},$curver);
	return ( $tables_new, $limits );
}

sub check_db_content($$$$) {
	my ( $dbh, $sql, $db, $tables ) = @_;
	my @queries = split /;\s*$/m, $sql;
	my $tmp;
	for my $q ( @queries ) {
		$q =~ s/^-- .*$//mg;
		$q =~ s/\/\*.*\*\///mg;
		$q =~ s/^\s*//s;
		$q =~ s/\s*$//s;
		next if ( $q =~ /^create table/i );
		next if ( $q =~ /^drop table/i );
		if ( $q =~ /^\s*replace into SEQ\s+values\s*\('(\w+)'\s*,(.*)\)\s*$/i ) {
			my $name = $1;
			my $seq_info = $2;
			my $seq  = get_current($dbh,$name);
			if ( ! defined $seq ) {
				if ( $dry_data ) {
					print "-- Need to create SEQ $name\n";
					print $q.";\n";
					next;
				}
				print "Creation SEQ $name\n";
				$tmp = sql_transaction($dbh,'replace',$q);
				if ( ! $tmp->{'ok'} ) {
					die "Cannot create SEQ on $name, in databse $db:".Dumper($tmp);
				}
				next;
			} else {
				# I may need to update the SEQ entry
				if ( $seq_info =~ /^\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*,.*$/ ) {
					my ( $current, $min, $max ) = ( $1, $2, $3 );
					$tmp = sql_transaction($dbh, 'select', "select * from SEQ where SEQ_NAME = '$name'");
					if ( ! $tmp->{'ok'} ) {
						die "Cannot read SEQ named $name, in database $db:".Dumper($tmp);
					}
					my $seq = $tmp->{'rows'}->[0];
					my $q2;
					if ( $seq->{'SEQ_MINVALUE'} ne $min ) {
						$q2 = "update SEQ set SEQ_MINVALUE = '$min' where SEQ_NAME = '$name'";
						if ( $dry_data ) {
							print "-- Need to update min value for sequence $name\n";
							print $q2.";\n";
						} else {
							print "Update SEQ_MINVALUE on SEQ $name to $min\n";
							$tmp = sql_transaction($dbh,'update',$q2);
							if ( ! $tmp->{'ok'} ) {
								die "Failed to update SEQ on $name, in database $db:".Dumper($tmp);
							}
						}
					}
					if ( $seq->{'SEQ_MAXVALUE'} ne $max ) {
						$q2 = "update SEQ set SEQ_MAXVALUE = '$max' where SEQ_NAME = '$name'";
						if ( $dry_data ) {
							print "-- Need to update max value for sequence $name\n";
							print $q2.";\n";
						} else {
							print "Update SEQ_MAXVALUE on SEQ $name to $max\n";
							$tmp = sql_transaction($dbh,'update',$q2);
							if ( ! $tmp->{'ok'} ) {
								die "Failed to update SEQ on $name, in database $db:".Dumper($tmp);
							}
						}
					}
					next;
				} else {
					die "Failed to read SEQ numerical info from <$seq_info> in <$q>";
				}
			}
		}
		if ( $q =~ /^\s*replace into SEQ/i ) {
			die "Failed to read SEQ info: $q\n";
		}
		next if ( $q !~ /^\s*insert\s+into\s+(`|)\w+\1\s+/i );
		if ( $q !~ /\s*insert\s+into\s+(`|)(\w+)\1\s+values/ixs ) {
			die "Cannot check insert: $q\n";
		}
		my $table = $2;
		my $vals  = [];
		my $t = {};
		$tmp = $q;
		if ( $tmp !~ s/^\s*insert\s+into\s+(`|)$table\1\s+values\s*\((.*)\)\s*$/$2/ixs) {
			die "Not an insert in $table: $q\n$tmp\n";
		}
		$tmp .= ',';
		for my $c ( @{$tables->{$table}->{'cols'}->{'list'}} ) {
			if ( $tmp !~ s/^\s*("|'|)(.*?)\1\s*,// ) {
				die "In table $table.\nSo far got ".Dumper($t)."\nCannot read after $tmp";
			}
			my ( $a, $b ) = ( $1, $2 );
			$t->{$c} = $2;
		}
		if ( $tmp ) {
			die "Too many values in $q\nRemains at the end <$tmp>\n";
		}

		if ( ! exists $t->{$table.'_ID'} ) {
			print "-- Je ne sais pas traiter les donnes des tables de relation (table $table)\n";
			next;
		}

		my $row = __par_id($dbh,$table,$t->{$table.'_ID'});
		if ( ! defined $row ) {
			if ( $dry_data ) {
				print "-- Missing line in the table\n";
				print $q.";\n";
				next;
			}
			print "Missing line id ".$t->{$table.'_ID'}." in table $db.$table\n";
			$tmp = sql_transaction($dbh,'insert',$q);
			if ( ! $tmp->{'ok'} ) {
				die "Failed with query $q\n".Dumper($tmp);
			}
			next;
		}

		my $diff = 0;
		for my $k ( keys %$t ) {
			next if ( ! defined $row->{$k} && $t->{$k} =~ /^NULL$/i );
			next if ( $t->{$k} eq 'now()' );
			if ( $tables->{$table}->{'cols'}->{$k}->{'type'} =~ /^bool(?:ean)?$/i ) {
				next if ( ! $row->{$k} && $t->{$k} =~ /^false$/i );
				next if ( $row->{$k} && $t->{$k} =~ /^true$/i );
				print "-- Differs on boolean value on $k: is ".$row->{$k}." should be ".$t->{$k}."\n";
				$diff = 1;
				next;
			}
			if ( ! defined $row->{$k} ) {
				print "-- Differs at least on $k: is NULL should be ".$t->{$k}."\n";
				$diff = 1;
				next;
			}
			if ( $row->{$k} ne $t->{$k} ) {
				print "-- Differs at least on $k: is ".$row->{$k}." should be ".$t->{$k}."\n";
				$diff = 1;
			}
		}
		if ( $diff ) {
			$q =~ s/insert/replace/s;
			if ( $dry_data ) {
				print $q.";\n";
			} else {
				print "Updating in table $table id ".$t->{$table.'_ID'}."\n";
				$tmp = sql_transaction($dbh,'replace',$q);
				if ( ! $tmp->{'ok'} ) {
					die "Failed to update line in table $table, id ".$t->{$table.'_ID'}." with $q:".Dumper($tmp);
				}
			}
		}

		my $seq = get_current($dbh,$table.'_ID');
		my $nseq;
		while ( $seq < $t->{$table.'_ID'} ) {
			if ( $dry_alter ) {
				print "-- Need to update SEQ table, no way to do that with a single request\n";
				$seq++;
			} else {
				print "Updating SEQ on $table: $seq < ".$t->{$table.'_ID'}."\n";
				$nseq = get_next($dbh,$table.'_ID');
				if ( ! $nseq ) {
					die "Cannot update SEQ table for ".$table.'_ID'."\n";
				}
				if ( $nseq <= $seq ) {
					die "SEQ did not increase during update for ".$table.'_ID'."\n";
				}
				$seq = $nseq;
			}
		}
	}
}

sub purge_db($$) {
	my ( $dbh, $db_name ) = @_;
	my ( $tmp, @tables );
	print "Purge content of $db_name\n";
	$tmp = sql_transaction($dbh,'show','show tables');
	if ( ! $tmp->{'ok'} ) {
		die "Failed to list tables for databse $db_name: ".Dumper($tmp)."\n";
	}
	return if ! defined $tmp->{'rows'};
	@tables = @{$tmp->{'rows'}};
	for my $t ( @tables ) {
		my $name = $t->{'Tables_in_'.$db_name};
		dry_or_run($dry_table, undef, $dbh, 'drop', "drop table $name");
	}
}

sub execute_queries($$) {
	my ( $dbh, $sql ) = @_;
	my @queries = split /;\s*$/m, $sql;
	my $tmp;
	for my $q ( @queries ) {
		next if $q =~ /^-- /;
		next if $q =~ /^\s*$/;
		my $text;
		if ( $q =~ /create\s+table\s+([a-z0-9_]+)\W/is ) {
			$text = " $1,";
		}
		dry_or_run($dry_table, $text, $dbh, 'create', $q);
	}
	print "\n" if ! $dry_table;
}


sub structure() {
	read_files;
	if ( ! $dbh ) {
		die "Failed to connect to $db_name: ".Dumper($dbh)."\n";
	}
	if ( $force ) {
		print "Building the database\n";
		purge_db($dbh, $db_name);
		print "Create $db_name ";
		execute_queries($dbh,$create);
	} else {
		print "-- Checking the database\n";
		if ( $dry_table || $dry_alter || $dry_data ) {
			print "-- Check database $db_name...\n";
			print "use $db_name;\n";
		}
		my $curver = get_current_version($dbh);
		parse_db_structure;
		check_db_structure;
		check_db_content($dbh,$create,$db_name,$tables_new);
		check_db_content($dbh,$data,$db_name,$tables_new);
		check_limits($dbh,$limits->{'post'},$curver);
	}
	disconnect_db($dbh);
}

##############################################################################################
# About the script itself (options and so on)
##############################################################################################

sub check_db(;$) {
	my ( $action ) = @_;
	@actions = ( $action ) if defined $action;

	for my $act ( @actions ) {
		print "Checking action $act\n";
		if ( ! control_options($act) ) {
			die "Cannot perform <$act> as requested";
		}

		print "Processing action $act\n";
		if ( $act eq 'structure' ) {
			structure();
			next;
		}
		if ( $act eq 'create' ) {
			create();
			next;
		}
		if ( $act eq 'config' ) {
			config();
			next;
		}
		die "Unknown action <$act> to perform";
	}
}

sub switch_option($$$) {
	my ( $groupe, $on, $select ) = @_;
	my $grps = {
		'dry' => {
			'all' => [ \$dry_db, \$dry_table, \$dry_alter, \$dry_data, \$dry_cfg ],
			'db' => [ \$dry_db ],
			'table' => [ \$dry_table ],
			'alter' => [ \$dry_alter ],
			'data' => [ \$dry_data ]
		}
	};
	die if ! defined $grps->{$groupe};
	die if ! defined $grps->{$groupe}->{$select};
	for my $opt ( @{$grps->{$groupe}->{$select}} ) {
		$$opt = $on;
	}
	return;
}

sub add_sql_file($$) {
	my ( $option, $file ) = @_;
	push @files, $file;
	__debug "Adding file $file\n";
}

sub set_db_name($$) {
	my ( $option, $db ) = @_;
	$db_name = $db;
	__debug "Switch to database $db\n";
}

sub set_db_force($$) {
	my ( $option, $f ) = @_;
	$force = $f;
}

sub set_login($$) {
	my ( $option, $l ) = @_;
	$login = $l;
}

sub set_passwd($$) {
	my ( $option, $p ) = @_;
	$passwd = $p;
}

sub set_utf8($$) {
	my ( $option, $u ) = @_;
	$utf8 = $8;
}

sub set_type($$) {
	my ( $option, $t ) = @_;
	die "The only possible type for now is mysql" if $t ne "mysql";
}

sub set_host($$) {
	my ( $option, $h ) = @_;
	$host = $h;
}

sub set_source($$) {
	my ( $option, $s ) = @_;
	$source = $s;
}

sub set_config_file($$) {
	my ( $option, $f ) = @_;
	$config_file = $f;
}

sub set_config($$) {
	my ( $option, $c ) = @_;
	my $cfg = Config_Section($config_file,$c);
	die "Config not found in /etc/fdn/$config_file" if ! defined $cfg;
	die "Only the 'mysql' type is supported so far" if defined $cfg->{'type'} && $cfg->{'type'} ne 'mysql';
	$db_name = $c;
	$host = $cfg->{'host'} if defined $cfg->{'host'};
	$source = $cfg->{'source'} if defined $cfg->{'source'};
	$db_name = $cfg->{'name'} if defined $cfg->{'name'};
	$login = $cfg->{'login'} if defined $cfg->{'login'};
	$passwd = $cfg->{'passwd'} if defined $cfg->{'passwd'};
	$utf8 = $cfg->{'utf8'} if defined $cfg->{'utf8'};
	push @files, @{$cfg->{'@files'}} if defined $cfg->{'@files'};
	push @actions, @{$cfg->{'@actions'}} if defined $cfg->{'@actions'};
}

sub control_options($) {
	my ( $action ) = @_;
	my $err = [];
	if ( $action eq 'structure' ) {
		$dbh = undef;
		if ( ! -r '/etc/fdn/dbc.cfg' ) {
			push @$err, "No access configured on that host";
		}
		if ( ! defined $db_name ) {
			push @$err, "No database selected";
		} else {
			my $cfg = Config_Section('dbc.cfg',$db_name);
			push @$err, "No configuration for database $db_name on that host (/etc/fdn/dbc.cfg)" if ! defined $cfg;
			eval { $dbh = do_connect($db_name) };
			push @$err, "Failed to connect to database $db_name on that host (/etc/fdn/dbc.cfg)" if ! defined $dbh;
		}
		push @$err, "No sql file selected" if $#files < 0;
		for my $f ( @files ) {
			if ( ! -e $f ) {
				push @$err, "File not found: $f";
				next;
			}
			if ( ! -r $f ) {
				push @$err, "Permission denied: $f";
			}
		}
	}
	if ( $action eq 'config' ) {
	}
	if ( $action eq 'create' ) {
		my $cfg = Config_Section('dbc.cfg','GLOBAL');
		push @$err, "No configuration for root db on that host (/etc/fdn/dbc.cfg)" if ! defined $cfg;
		$dbh = do_connect('GLOBAL');
		push @$err, "Failed to connect as root ([GLOBAL] in /etc/fdn/dbc.cfg)" if ! defined $dbh;
		push @$err, "No database selected" if ! defined $db_name;
		push @$err, "No login provided" if ! defined $login;
	}
	if ( $#$err >= 0 ) {
		$err = join "\n", @$err;
		print STDERR $err."\n";
		return 0;
	}
	return 1;
}


#############################################################################################
# To check the texts in the localisation database.
#############################################################################################

sub check_text($$$$) {
	my ( $dbh, $appli, $aid, $t ) = @_;
	#ID SIMPLE DESC VAL
	#print "Je vrifie le texte $appli/$aid/".$t->{'TEXT_ID'}." (".$t->{'TEXT_VAL'}.")\n";
	my $text = &$text_par_xid_appli_id($t->{'TEXT_ID'},$aid);
	if ( defined $text ) {
		if ( $text->{'TEXT_VAL'} ne $t->{'TEXT_VAL'} ) {
			&$set_text_val($text->{'TEXT_ID'},$t->{'TEXT_VAL'});
		}
		if ( $text->{'TEXT_SIMPLE'} ne $t->{'TEXT_SIMPLE'} ) {
			&$set_text_simple($text->{'TEXT_ID'},$t->{'TEXT_SIMPLE'});
		}
	} else {
		print "Il faut crer le texte, application: $appli/$aid, ".$t->{'TEXT_VAL'}."\n";
		my $tid = &$create_text($aid, $t->{'TEXT_ID'}, $t->{'TEXT_VAL'}, $t->{'TEXT_SIMPLE'}, $t->{'TEXT_DESC'});
		if ( ! $tid ) {
			die "Failed to create text\n";
		}
	}
}

sub check_texts($$) {
	my ( $dbh, $appli ) = @_;
	my $sql = $texts->{$appli};
	my @queries = split /;\s*$/m, $sql;
	my $aid = &$appli_id_par_name($appli);
	if ( ! defined $aid ) {
		print "-- Failed to get application $appli, need to insert it\n";
		if ( ! $dry_data ) {
			$aid = &$new_appli($appli);
		}
		if ( ! $aid ) {
			# FIXME If in dry run, normal failure (ask for --run)
			die "Failed to insert row in table APPLICATION\n";
		}
	}
	print "-- On application '$appli' / <$aid>\n";
	for my $q ( @queries ) {
		next if ( $q !~ /^\s*insert into TEXT/ );
		my $t = {};
		if ( $q !~ /\s*insert\s+into\s+TEXT\s+
				\( \s*(\w+)\s*,   \s*(\w+)\s*,   \s*(\w+)\s*,   \s*(\w+)\s*  \)
				\s* values \s*
				\( \s*"((?:[^"]|\\")*)"\s*,  \s*"((?:[^"]|\\")*)"\s*,  \s*"((?:[^"]|\\")*)"\s*,  \s*"((?:[^"]|\\")*)"\s* \)\s*$/xs ) {
			die "Invalid insert in text database for appli $appli: $q\n";
		}
		$t->{$1} = $5;
		$t->{$2} = $6;
		$t->{$3} = $7;
		$t->{$4} = $8;
		$t->{'TEXT_VAL'} =~ s/\\"/"/gs;
		check_text( $dbh, $appli, $aid, $t );
	}
}

sub check_all_the_texts() {
	my $login = $connect->{'localisation'}->[0];
	my $passwd = $connect->{'localisation'}->[1];
	my $opts = { 'PrintError' => 0 };
	$opts->{'utf8'} = $connect->{'localisation'}->[2] if defined $connect->{'localisation'}->[2];
	my $tmp;
	my $dbh = connect_db('mysql','localhost','localisation',$login,$passwd,$opts);
	if ( ! $dbh ) {
		die "Failed to connect to localisation databse\n";
	}
	print "-- Checking the texts\n";
	for my $appli ( keys %$texts ) {
		print
		check_texts($dbh,$appli);
	}
	disconnect_db($dbh);
}

sub utf8_clean($) {
	my ( $hash ) = @_;
	for my $k ( %$hash ) {
		Encode::_utf8_on($hash->{$k});
	}
}

1;
